"""Module for usefull tools."""
from os import remove
from os.path import exists

import numpy as np
import periodictable as pt
from ase import Atoms
from ase.calculators.calculator import Calculator
from ase.db import connect
from ase.io.trajectory import Trajectory

from flare import otf_parser

def traj_to_db(traj_file, db_file, append=False):
    """Convert ASE trajectory file to database.

    Additionally, the tag 'relaxed' is set to False.

    Args:
        traj_file (:obj:`Path`): trajectory filename.
        db_file (:obj:`Path`): database filename.
        append (:obj:`bool`): append to a current database if it exists.
            If database file does not exists this does nothing.
    """
    # File Checks
    if not exists(traj_file):
        raise FileNotFoundError(traj_file + " is not in this directory!")
    if exists(db_file) and append:
        print(db_file + " already exists. Appending to it...")
    elif exists(db_file) and not append:
        print(db_file + " already exists. Overwriting...")
        remove(db_file)

    # Open files
    database = connect(db_file)
    traj = Trajectory(traj_file)

    # Write to database
    for atoms in traj:
        database.write(atoms, relaxed=False)

def otf_to_db(otf_file, db_file, append=False):
    """Convert FLARE OTF file to database.

    Additionally, the tag 'relaxed' is set to False.

    Args:
        otf_file (:obj:`Path`): OTF filename.
        db_file (:obj:`Path`): database filename.
        append (:obj:`bool`): append to a current database if it exists.
            If database file does not exists this does nothing.
    """
    # File checks
    if not exists(otf_file):
        raise FileNotFoundError(otf_file + " is not in this directory!")
    if exists(db_file) and append:
        print(db_file + " already exists. Appending to it...")
    elif exists(db_file) and not append:
        print(db_file + " already exists. Overwriting...")
        remove(db_file)

    # Open files
    database = connect(db_file)
    otf = otf_parser.OtfAnalysis(otf_file, calculate_energy=True)

    types = otf.header['species']
    atomicnumbers = np.zeros(len(types))
    atomicmasses = np.zeros(len(types))
    for i, _ in enumerate(types):
        atomicnumbers[i] = getattr(pt, types[i]).number
        atomicmasses[i] = getattr(pt, types[i]).mass

    for i, _ in enumerate(otf.dft_frames):
        # Get all the interesting data
        forces = otf.gp_force_list[i]
        positions = otf.gp_position_list[i]
        energy = otf.gp_thermostat['potential energy'][i]

        # Set all the properties
        atoms = Atoms(''.join(types), positions=positions)
        atoms.calc = Calculator()
        atoms.calc.implemented_properties = ['forces', 'energy']
        atoms.calc.results['forces'] = forces
        atoms.calc.results['energy'] = energy
        atoms.numbers = atomicnumbers
        atoms.masses = atomicmasses

        database.write(atoms)



"""Module containing some utilities for testing."""

from pathlib import Path
import pkg_resources as pkg

__all__ = ["PATH_QMCBLIP", "PATH_TEST"]

# Environment data
PATH_QMCBLIP = Path(pkg.resource_filename('qmcblip', ''))
ROOT = PATH_QMCBLIP.parent

PATH_TEST = ROOT / "tests"



import unittest
from qmcblip.champio import Settings
from pydantic import ValidationError

class TestChampio(unittest.TestCase):
    def test_SetupError(self):
        with self.assertRaises(ValidationError):
            Settings()


if __name__ == '__main__':
    unittest.main()


"""Custom Verlet scheme."""
import numpy as np
from ase.md.verlet import VelocityVerlet

class CustomVerlet(VelocityVerlet):
    """Custom Verlet scheme for ASE and FLARE.
    """
    def __init__(self, atoms, timestep=None, trajectory=None, logfile=None,
                 loginterval=1, dt=None, append_trajectory=False):

        VelocityVerlet.__init__(self, atoms, timestep, trajectory, logfile,
                                   loginterval, dt,
                                   append_trajectory=append_trajectory)
        self.old_forces = None

    def step(self, forces=None):

        atoms = self.atoms

        if forces is None:
            forces = atoms.get_forces(md=True)

        p = atoms.get_momenta()

        if self.old_forces is None:
            p = p
            #p += 0.5 * self.dt * forces
        else:
            p += 0.5 * self.dt * (forces + self.old_forces)

        atoms.set_momenta(p)

        masses = atoms.get_masses()[:, np.newaxis]
        r = atoms.get_positions()

        atoms.set_positions(r + self.dt * p / masses + 0.5 * self.dt**2 * forces / masses)

        self.old_forces = forces

        return forces



"""ASE CHAMPS Calculator."""
from os import path

import numpy as np
from ase.calculators.calculator import (CalculatorSetupError, FileIOCalculator,
                                        InputError)
from ase.units import Ang, Bohr, Ha

from qmcblip.champio import Settings


class CHAMP(FileIOCalculator):
    """ASE Calculator for CHAMP.

    An ASE calculator for the Quantum Monte Carlo software CHAMP.

    Args:
        vmc_in (:obj:`str`, optional): File to read CHAMP settings from.
        vmc_out (:obj:`str`, optional):The output file of CHAMP.
        force_file (:obj:`str`, optional): The file that CHAMP writes the forces and energies to.
        pos_file (:obj:`str`, optional): The file from which CHAMP will read to
            location of the atoms.
        champ_loc (:obj:`str`, optional): Location of the CHAMP executable.
        nodefile (:obj:`str`, optional): If set, the calculator will run on
            multiple nodes for CHAMP.
        ncore (:obj:`str`, optional): Amount of cores to run CHAMP on.
        settings (:obj:`Settings<qmcblip.champio.Settings>`, optional): Input settings for CHAMP
            (OVERRULES VMC.INP).
        use_opt_wf (:obj:`bool`, optional): Use the optimized WF from last step.
    """

    # Stress only for FLARE compatibility
    implemented_properties = ['energy', 'forces', 'stress']

    name = "CHAMP"

    # The parameters to run CHAMP
    default_parameters = dict(
        vmc_in='vmc.inp',
        vmc_out='vmc.out',
        force_file='write_forces',
        pos_file='molecule.xyz',
        champ_loc='/usr/bin/vmc.mov1',
        nodefile=None,
        ncore=1,
        settings=None,
        use_opt_wf=False)

    # Placeholder
    command = ""

    def __init__(self, restart=None,
                ignore_bad_restart_file=FileIOCalculator._deprecated,
                label='CHAMP', atoms=None, **kwargs):

        FileIOCalculator.__init__(self, restart, ignore_bad_restart_file,
                                label, atoms, **kwargs)

        if not path.exists(self.parameters['champ_loc']):
            raise CalculatorSetupError("Did not find champ at: " + self.parameters['champ_loc'])

        if (not path.exists(self.parameters['vmc_in']) and self.parameters['settings'] is None):
            raise CalculatorSetupError("You did no supply any configuration parameters for CHAMP. \
                                       Either give a .inp file or give a configuration dictionary.")

        if (not isinstance(self.parameters['settings'], Settings) and
            self.parameters['settings'] is not None):
            raise CalculatorSetupError('You did not provide a proper settings object.')

        if self.parameters['settings'] is not None:
            self.parameters['settings'].write(filename=self.parameters['vmc_in'])
        else:
            self.parameters['settings'] = Settings.read(self.parameters['vmc_in'])

        # Set the command to call CHAMP
        self._set_command()

    def _set_command(self):
        # Add nodefile or not
        if self.parameters['nodefile'] is None:
            self.command = "mpirun -n " + str(self.parameters['ncore']) + " " \
                    + self.parameters['champ_loc'] + " -i vmc_temp.inp" \
                    + " -o " + self.parameters['vmc_out']
        else:
            self.command = "mpirun -s all -n " + str(self.parameters['ncore']) \
                    + " -machinefile " + self.parameters['nodefile'] + " " \
                    + self.parameters['champ_loc'] + " -i vmc_temp.inp" \
                    + " -o " + self.parameters['vmc_out']

    def configure(self, **kwargs):
        """(Re)configure the CHAMP calculator by setting the keyword arguments.

        The tags for the vmc.inp can also be set here.

        Args:
            **kwargs: keyword arguments.
        """
        self.set(**kwargs)

        # Set the command to call CHAMP. May have changed due to the parameters.
        self._set_command()

    def write_input(self, atoms, properties=None, system_changes=None):
        FileIOCalculator.write_input(self, atoms, properties, system_changes)

        # Write the coordinates of the atoms in xyz format
        with open(self.parameters['pos_file'], 'w', encoding='utf-8') as fileobj:
            comment = atoms.symbols
            natoms = len(atoms)
            fmt='%22.15f'

            fileobj.write('%d\n%s\n' % (natoms, comment))

            for label, (x, y, z) in zip(atoms.symbols, atoms.positions):
                fileobj.write('%-2s %s %s %s\n' % (label, fmt % (x * Ang/Bohr), \
                                fmt % (y * Ang/Bohr), fmt % (z * Ang/Bohr)))

        if self.parameters['use_opt_wf']:
            self.parameters['settings'].use_opt_wf()

        # Rewrite the input file
        self.parameters['settings'].write('vmc_temp.inp')



    def read(self, label):
        raise NotImplementedError

    def read_results(self):
        if not path.exists(self.parameters['force_file']):
            raise InputError("Could not detect the forces from CHAMP in " \
                                        + self.parameters['force_file'])

        with open(self.parameters['force_file'], 'r', encoding="utf-8") as fileobj:
            # Read the energy and convert to eV
            self.results['energy'] = float(fileobj.readline()) * Ha

            # Read the forces
            forces = []
            for line in fileobj.readlines():
                f_list = [float(i) for i in line.split(" ") if i.strip()]
                forces.append(f_list)

            # Convert to eV/A and use ASE force direction
            self.results['forces'] = -np.asarray(forces) * Ha/Bohr

        # Pass an null stress tensor for FLARE compatibility
        self.results['stress'] = np.zeros((6,))



"""Tools for simulations with sGDML."""
import os
from os.path import exists

import numpy as np
from ase.io import read

from sgdml import __version__
from sgdml.utils import io

def db_to_sgdml(db_file, dataset_file, name=None):
    """Convert an ASE DB to a sGDML dataset.

    Note:
        Licensed under MIT license.
        Copyright (c) 2018-2020 Stefan Chmiela

    Args:
        db_file (:obj:`Path`): database filename.
        dataset_file (:obj:`Path`): sGDML dataset file.
        name (:obj:`str`): name of the new dataset.
    """

    # MIT License
    #
    # Copyright (c) 2018-2020 Stefan Chmiela
    #
    # Permission is hereby granted, free of charge, to any person obtaining a copy
    # of this software and associated documentation files (the "Software"), to deal
    # in the Software without restriction, including without limitation the rights
    # to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    # copies of the Software, and to permit persons to whom the Software is
    # furnished to do so, subject to the following conditions:
    #
    # The above copyright notice and this permission notice shall be included in all
    # copies or substantial portions of the Software.
    #
    # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    # SOFTWARE.

    # File checks
    if not exists(db_file):
        raise FileNotFoundError(db_file + " is not in this directory!")
    if exists(dataset_file):
        print(dataset_file + " already exists. Overwriting...")

    basename = os.path.splitext(os.path.basename(db_file))[0]

    mols = read(db_file, index=':')

    # filter incomplete outputs from trajectory
    mols = [mol for mol in mols if mol.get_calculator() is not None]

    R, z, E, F = None, None, None, None

    calc = mols[0].get_calculator()

    if 'forces' not in calc.results:
        raise ValueError("Forces are missing in the database file!")

    Z = np.array([mol.get_atomic_numbers() for mol in mols])
    if not (Z == Z[0]).all():
        raise ValueError('Order of atoms changes accross database!')

    R = np.array([mol.get_positions() for mol in mols])
    z = Z[0]

    E = np.array([mol.get_potential_energy() for mol in mols])
    F = np.array([mol.get_forces() for mol in mols])

    if name is None:
        name = basename

    # Base variables contained in every model file.
    base_vars = {
        'type': 'd',
        'code_version': __version__,
        'name': name,
        'theory': 'QMC',
        'R': R,
        'z': z,
        'F': F,
    }

    base_vars['F_min'], base_vars['F_max'] = np.min(F.ravel()), np.max(F.ravel())
    base_vars['F_mean'], base_vars['F_var'] = np.mean(F.ravel()), np.var(F.ravel())

    base_vars['r_unit'] = 'Ang'
    base_vars['e_unit'] = 'eV'

    if E is not None:
        base_vars['E'] = E
        base_vars['E_min'], base_vars['E_max'] = np.min(E), np.max(E)
        base_vars['E_mean'], base_vars['E_var'] = np.mean(E), np.var(E)

    base_vars['md5'] = io.dataset_md5(base_vars)
    np.savez_compressed(dataset_file, **base_vars)



"""On-the-fly training classes adapted for CHAMP."""
# MIT License

# Copyright (c) 2019 Harvard University

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE

from time import time
from copy import deepcopy
import numpy as np

from flare.otf import OTF
from flare.ase.otf import ASE_OTF
from flare.ase.atoms import FLARE_Atoms
import flare.ase.dft as dft_source
from flare.utils.learner import is_std_in_bound

from ase import units

from .verlet import CustomVerlet

class C_OTF(OTF):
    """Trains a Gaussian process force field on the fly during molecular dynamics.

    Args:
        dt (float): MD timestep.
        number_of_steps (:obj:`int`): Number of timesteps in the training
            simulation.
        prev_pos_init ([type], optional): Previous positions. Defaults
            to None.
        rescale_steps (List[:obj:`int`], optional): List of frames for which the
            velocities of the atoms are rescaled. Defaults to [].
        rescale_temps (List[:obj:`int`], optional): List of rescaled temperatures.
            Defaults to [].
        gp (gp.GaussianProcess): Initial GP model.
        calculate_energy (:obj:`bool`, optional): If True, the energy of each
            frame is calculated with the GP. Defaults to False.
        calculate_efs (:obj:`bool`, optional): If True, the energy and stress of each
            frame is calculated with the GP. Defaults to False.
        write_model (:obj:`int`, optional): If 0, write never. If 1, write at
            end of run. If 2, write after each training and end of run.
            If 3, write after each time atoms are added and end of run.
            If 4, write after each training and end of run, and back up
            after each write.
        force_only (:obj:`bool`, optional): If True, only use forces for training.
            Default to False, use forces, energy and stress for training.
        std_tolerance_factor (float, optional): Threshold that determines
            when DFT is called. Specifies a multiple of the current noise
            hyperparameter. If the epistemic uncertainty on a force
            component exceeds this value, DFT is called. Defaults to 1.
        skip (:obj:`int`, optional): Number of frames that are skipped when
            dumping to the output file. Defaults to 0.
        init_atoms (List[:obj:`int`], optional): List of atoms from the input
            structure whose local environments and force components are
            used to train the initial GP model. If None is specified, all
            atoms are used to train the initial GP. Defaults to None.
        output_name (:obj:`str`, optional): Name of the output file. Defaults to
            'otf_run'.
        max_atoms_added (:obj:`int`, optional): Number of atoms added each time
            DFT is called. Defaults to 1.
        freeze_hyps (:obj:`int`, optional): Specifies the number of times the
            hyperparameters of the GP are optimized. After this many
            updates to the GP, the hyperparameters are frozen.
            Defaults to 10.
        min_steps_with_model (:obj:`int`, optional): Minimum number of steps the
            model takes in between calls to DFT. Defaults to 0.
        force_source (Union[:obj:`str`, object], optional): DFT code used to calculate
            ab initio forces during training. A custom module can be used here
            in place of the DFT modules available in the FLARE package. The
            module must contain two functions: parse_dft_input, which takes a
            file name (in string format) as input and returns the positions,
            species, cell, and masses of a structure of atoms; and run_dft_par,
            which takes a number of DFT related inputs and returns the forces
            on all atoms.  Defaults to "qe".
        npool (:obj:`int`, optional): Number of k-point pools for DFT
            calculations. Defaults to None.
        mpi (:obj:`str`, optional): Determines how mpi is called. Defaults to
            "srun".
        dft_loc (:obj:`str`): Location of DFT executable.
        dft_input (:obj:`str`): Input file.
        dft_output (:obj:`str`): Output file.
        dft_kwargs ([type], optional): Additional arguments which are
            passed when DFT is called; keyword arguments vary based on the
            program (e.g. ESPRESSO vs. VASP). Defaults to None.
        store_dft_output (Tuple[Union[:obj:`str`,List[:obj:`str`]],:obj:`str`], optional):
            After DFT calculations are called, copy the file or files
            specified in the first element of the tuple to a directory
            specified as the second element of the tuple.
            Useful when DFT calculations are expensive and want to be kept
            for later use. The first element of the tuple can either be a
            single file name, or a list of several. Copied files will be
            prepended with the date and time with the format
            'Year.Month.Day:Hour:Minute:Second:'.
        n_cpus (:obj:`int`, optional): Number of cpus used during training.
            Defaults to 1.
    """

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

    def run(self):
        """Performs an on-the-fly training run.

        If OTF has store_dft_output set, then the specified DFT files will
        be copied with the current date and time prepended in the format
        'Year.Month.Day:Hour:Minute:Second:'.
        """

        optional_dict = {"Restart": self.curr_step}
        self.output.write_header(
            str(self.gp),
            self.dt,
            self.number_of_steps,
            self.structure,
            self.std_tolerance,
            optional_dict,
        )

        counter = 0
        self.start_time = time()

        while self.curr_step < self.number_of_steps:
            # run DFT and train initial model if first step and DFT is on
            if (
                (self.curr_step == 0)
                and (self.std_tolerance != 0)
                and (len(self.gp.training_data) == 0)
            ):

                # Are the recorded forces from the GP or DFT in ASE OTF?
                # When DFT is called, ASE energy, forces, and stresses should
                # get updated.
                self.initialize_train()

            # after step 1, try predicting with GP model
            else:
                # compute forces and stds with GP
                self.dft_step = False
                self.compute_properties()

                # get max uncertainty atoms
                std_in_bound, target_atoms = is_std_in_bound(
                    self.std_tolerance,
                    self.gp.force_noise,
                    self.structure,
                    max_atoms_added=self.max_atoms_added,
                    update_style=self.update_style,
                    update_threshold=self.update_threshold,
                )

                steps_since_dft = self.curr_step - self.last_dft_step
                if (not std_in_bound) and (steps_since_dft > self.min_steps_with_model):
                    # record GP forces
                    self.update_temperature()
                    self.record_state()
                    gp_energy = self.structure.potential_energy
                    gp_forces = deepcopy(self.structure.forces)
                    gp_stress = deepcopy(self.structure.stress)

                    # run DFT and record forces
                    self.dft_step = True
                    self.last_dft_step = self.curr_step
                    self.run_dft()
                    dft_frcs = deepcopy(self.structure.forces)
                    dft_stress = deepcopy(self.structure.stress)
                    dft_energy = self.structure.potential_energy

                    # run MD step & record the state
                    self.record_state()

                    # record DFT data into an .xyz file with filename self.dft_xyz.
                    # the file includes the structure, e/f/s labels and atomic
                    # indices of environments added to gp
                    self.record_dft_data(self.structure, target_atoms)

                    # compute mae and write to output
                    self.compute_mae(
                        gp_energy,
                        gp_forces,
                        gp_stress,
                        dft_energy,
                        dft_frcs,
                        dft_stress,
                    )

                    # add max uncertainty atoms to training set
                    self.update_gp(
                        target_atoms,
                        dft_frcs,
                        dft_stress=dft_stress,
                        dft_energy=dft_energy,
                    )

                    if self.write_model == 4:
                        self.checkpoint()
                        self.backup_checkpoint()

            # write gp forces
            if counter >= self.skip and not self.dft_step:
                self.update_temperature()
                self.record_state()
                counter = 0

            counter += 1

            self.md_step(forces=self.structure.forces)  # update positions by Verlet
            self.rescale_temperature(self.structure.positions)

            if self.write_model == 3:
                self.checkpoint()

        self.output.conclude_run()

        if self.write_model >= 1:
            self.write_gp()
            self.checkpoint()

    def initialize_train(self):
        # call dft and update positions
        self.run_dft()
        dft_frcs = deepcopy(self.structure.forces)
        dft_stress = deepcopy(self.structure.stress)
        dft_energy = self.structure.potential_energy

        self.update_temperature()
        self.record_state()
        self.record_dft_data(self.structure, self.init_atoms)

        # make initial gp model and predict forces
        self.update_gp(
            self.init_atoms, dft_frcs, dft_stress=dft_stress, dft_energy=dft_energy
        )
        self.structure.forces = dft_frcs

class C_ASE_OTF(ASE_OTF, C_OTF):
    """On-the-fly training module using ASE MD engine, a subclass of OTF.

    Args:
        atoms (ASE Atoms): the ASE Atoms object for the on-the-fly MD run.
        timestep (:obj:`float`): the timestep in MD. Please use ASE units, e.g. if the
            timestep is 1 fs, then set `timestep = 1 * units.fs`
        number_of_steps (:obj:`int`): the total number of steps for MD.
        dft_calc (ASE Calculator): any ASE calculator is supported,
            e.g. Espresso, VASP etc.
        md_engine (:obj:`str`): the name of MD thermostat, only `VelocityVerlet`,
            `NVTBerendsen`, `NPTBerendsen`, `NPT` and `Langevin`, `NoseHoover`
            are supported.
        md_kwargs (dict): specify the args for MD as a dictionary, the args are
            as required by the ASE MD modules consistent with the `md_engine`.
        update_settings (List[List[dict]]): array containg CHAMP simulation
            parameters to update.
        calculator (:obj:`Calculator <ase.calculators.calculator.Calculator>`): ASE calculator.
            Must have "get_uncertainties" method
          implemented.
        trajectory (ASE Trajectory): default `None`, not recommended,
            currently in experiment.

    The following arguments are for on-the-fly training, the user can also
    refer to :class:`flare.otf.OTF`

    Args:
        prev_pos_init ([type], optional): Previous positions. Defaults
            to None.
        rescale_steps (List[:obj:`int`], optional): List of frames for which the
            velocities of the atoms are rescaled. Defaults to [].
        rescale_temps (List[:obj:`int`], optional): List of rescaled temperatures.
            Defaults to [].
        calculate_energy (:obj:`bool`, optional): If True, the energy of each
            frame is calculated with the GP. Defaults to False.
        write_model (:obj:`int`, optional): If 0, write never. If 1, write at
            end of run. If 2, write after each training and end of run.
            If 3, write after each time atoms are added and end of run.
            If 4, write after each training and end of run, and back up
            after each write.
        std_tolerance_factor (float, optional): Threshold that determines
            when DFT is called. Specifies a multiple of the current noise
            hyperparameter. If the epistemic uncertainty on a force
            component exceeds this value, DFT is called. Defaults to 1.
        skip (:obj:`int`, optional): Number of frames that are skipped when
            dumping to the output file. Defaults to 0.
        init_atoms (List[:obj:`int`], optional): List of atoms from the input
            structure whose local environments and force components are
            used to train the initial GP model. If None is specified, all
            atoms are used to train the initial GP. Defaults to None.
        output_name (:obj:`str`, optional): Name of the output file. Defaults to
            'otf_run'.
        max_atoms_added (:obj:`int`, optional): Number of atoms added each time
            DFT is called. Defaults to 1.
        freeze_hyps (:obj:`int`, optional): Specifies the number of times the
            hyperparameters of the GP are optimized. After this many
            updates to the GP, the hyperparameters are frozen.
            Defaults to 10.
        n_cpus (:obj:`int`, optional): Number of cpus used during training.
            Defaults to 1.
    """

    def __init__(
        self,
        atoms,
        timestep,
        number_of_steps,
        dft_calc,
        md_engine,
        md_kwargs,
        update_settings,
        calculator=None,
        trajectory=None,
        **otf_kwargs
    ):

        self.structure = FLARE_Atoms.from_ase_atoms(atoms)
        if calculator is not None:
            self.structure.calc = calculator
        self.timestep = timestep
        self.md_engine = md_engine
        self.md_kwargs = md_kwargs
        self._kernels = None
        self.update_settings = np.array(update_settings)

        if md_engine == "CustomVerlet":
            MD = CustomVerlet
        else:
            raise NotImplementedError(md_engine + " is not implemented in ASE")

        self.md = MD(
            atoms=self.structure,
            timestep=timestep,
            trajectory=trajectory,
            **md_kwargs,
        )

        force_source = dft_source
        self.flare_calc = self.structure.calc

        # Convert ASE timestep to ps for the output file.
        flare_dt = timestep / (units.fs * 1e3)

        C_OTF.__init__(
            self,
            dt=flare_dt,
            number_of_steps=number_of_steps,
            gp=self.flare_calc.gp_model,
            force_source=force_source,
            dft_loc=dft_calc,
            dft_input=self.structure,
            **otf_kwargs,
        )

        self.flare_name = self.output_name + "_flare.json"
        self.dft_name = self.output_name + "_dft.pickle"
        self.structure_name = self.output_name + "_atoms.json"
        self.checkpt_files = [
            self.checkpt_name,
            self.flare_name,
            self.dft_name,
            self.structure_name,
            self.dft_xyz,
        ]

    def md_step(self, forces=None):
        """Perform an MD step.

        Get new position in molecular dynamics based on the forces predicted by
        FLARE_Calculator or DFT calculator

        Args:
            forces (List[float]): array containing the forces as calculated by CHAMP (or FLARE).
        """

        if self.curr_step in self.update_settings and self.dtf_loc.name == "CHAMP":
            sets = self.dft_loc.parameters.settings
            ind = np.where(self.update_settings == self.curr_step)[0][0]
            changes = self.update_settings[ind][1]
            for key, val in changes.items():
                if isinstance(val, dict):
                    for key2, val2 in val.items():
                        setattr(getattr(sets, key), key2, val2)
                else:
                    setattr(sets, key, val)

        # Update previous positions.
        self.structure.prev_positions = np.copy(self.structure.positions)

        # Reset FLARE calculator.
        if self.dft_step:
            self.flare_calc.reset()
            self.structure.calc = self.flare_calc

        # Take MD step.
        # Inside the step() function, get_forces() is called
        self.md.step(forces=forces)
        self.curr_step += 1



"""Quickly run FLARE simulations."""
from typing import Any, List

import flare_pp._C_flare as flare_pp
from ase import units
from flare_pp.sparse_gp import SGP_Wrapper
from flare_pp.sparse_gp_calculator import SGP_Calculator
from pydantic import BaseModel

from flare.ase.calculator import FLARE_Calculator
from flare.gp import GaussianProcess
from flare.utils.parameter_helper import ParameterHelper

from .otf import C_ASE_OTF as ASE_OTF


class OTFSettings(BaseModel):
    """Dataclass containing FLARE(++) configuration.
    """

    class Theory(BaseModel):
        """Abstract baseclass.
        """

    class FLARE(Theory):
        """FLARE dataclass.
        """

        kernels: List[str] = ['twobody', 'threebody']
        """:obj:`List[str]`, optional: n-body functions to use."""

        cutoffs: List[float] = [5.0, 3.5]
        """:obj:`List[float]`, optional: cutoff for the n-body functions."""

        random: bool = True
        """:obj:`bool`, optional: randomize hyperparameters."""

        hyp_labels: List[str] = ['sig2','ls2','sig3','ls3','noise']
        """:obj:`List[str]`, optional: hyperparameter labels."""

        opt_algorithm: str = 'L-BFGS-B'
        """:obj:`str`, optional: hyperparamter optimization algorithm."""

        n_cpu: int = 1
        """:obj:`int`, optional: amount of CPU cores to run FLARE on."""

        update_style: str = "add_n"
        """:obj:`str`, optional: update algorithm for GP."""

        update_threshold: float = None
        """:obj:`str`, optional: update threshold for GP, in eV/Ang."""

        gp_model: Any = None
        flare_calc: Any = None

        def get_calc(self, atoms):
            """Make a FLARE calculator object.

            Args:
                atoms (:obj:`Atoms <ase.Atoms>`): atoms object.

            Returns:
                :obj:`FLARE_Calculator <flare.ase.calculator.FLARE_Calculator>`:
                    FLARE calculator object.
            """
            parameters = {}
            for ind, nbody in enumerate(self.kernels):
                parameters['cutoff_'+nbody] = self.cutoffs[ind]
            pm = ParameterHelper(
                kernels = self.kernels,
                random = self.random,
                parameters=parameters
            )

            hm = pm.as_dict()
            hyps = hm['hyps']
            cut = hm['cutoffs']

            if self.n_cpu > 1:
                parallel = True
            else:
                parallel = False

            self.gp_model = GaussianProcess(
                kernels = self.kernels,
                component = 'mc',
                hyps = hyps,
                cutoffs = cut,
                hyp_labels = self.hyp_labels,
                opt_algorithm = self.opt_algorithm,
                parallel = parallel,
                n_cpus = self.n_cpu
            )

            self.flare_calc = FLARE_Calculator(self.gp_model,
                                    par = True,
                                    mgp_model = None,
                                    use_mapping = False)
            return self.flare_calc


    class FLAREPP(Theory):
        """FLARE++ dataclass.
        """

        update_style: str = "threshold"
        """:obj:`str`, optional: update algorithm for GP."""

        update_threshold: float = 0.005
        """:obj:`str`, optional: update threshold for GP, in eV/Ang."""

        opt_algorithm: str = 'L-BFGS-B'
        """:obj:`str`, optional: hyperparamter optimization algorithm."""

        max_iterations: int = 10
        """:obj:`int`, optional: maximum number of hyperparameter optimization
            steps per MD step."""

        variance_type: str = 'local'
        """:obj:`str`, optional: uncertainty type on energy."""

        sigma_e: float = 0.12
        """:obj:`float`, optional: energy noise per atom, in kcal/mol."""

        sigma_f: float = 0.115
        """:obj:`float`, optional: force noise, in kcal/mol/Ang."""

        sigma_s: float = 0.014
        """:obj:`float`, optional: stress noise, in kcal/Ang^3."""

        cutoff: float = 5.0
        """:obj:`float`, optional: cutoff of kernel in Ang."""

        sigma: float = 2.0
        """:obj:`float`, optional: kernel thing in eV."""

        power: int = 2
        """:obj:`int`, optional: power of the kernel."""

        cutoff_function: str = "quadratic"
        """:obj:`str`, optional: cutoff function."""

        radial_basis: str = "chebyshev"
        """:obj:`str`, optional: radial basis."""

        N: int = 12
        """:obj:`int`, optional: number of radial basis functions."""

        lmax: int = 3
        """:obj:`int`, optional: largest L included in spherical harmonics."""

        kernel: Any = None
        descriptor_calculator: Any = None
        gp_model: Any = None
        flare_calc: Any = None

        def get_calc(self, atoms):
            """Make a FLARE++ calculator object.

            Args:
                atoms (:obj:`Atoms <ase.Atoms>`): atoms object.

            Returns:
                :obj:`SGP_Calculator <flare_pp.sparse_gp_calculator.SGP_Calculator>`:
                    FLARE++ calculator object.
            """
            species_map = {}
            for ind, num in enumerate(set(atoms.symbols.numbers)):
                species_map[num] = ind
            self.kernel = flare_pp.NormalizedDotProduct(self.sigma, self.power)
            radial_hyps = [0., self.cutoff]
            cutoff_hyps = []
            n_species = len(species_map)
            descriptor_settings = [n_species, self.N, self.lmax]
            self.descriptor_calculator = flare_pp.B2(
                                                self.radial_basis,
                                                self.cutoff_function,
                                                radial_hyps,
                                                cutoff_hyps,
                                                descriptor_settings
                                                )

            sigma_e = self.sigma_e * len(atoms)
            sigma_f = self.sigma_f
            sigma_s = self.sigma_s

            bounds = [(None, None), (sigma_e, None), (None, None), (None, None)]

            self.gp_model = SGP_Wrapper([self.kernel], [self.descriptor_calculator], self.cutoff,
                                sigma_e, sigma_f, sigma_s, species_map,
                                variance_type=self.variance_type,
                                stress_training=False,
                                opt_method=self.opt_algorithm,
                                bounds=bounds,
                                max_iterations=self.max_iterations)

            self.flare_calc = SGP_Calculator(self.gp_model)

            return self.flare_calc



    theory: Theory = FLARE()
    """:obj:`Theory`: FLARE or FLAREPP."""

    output_name: str = 'OTF'
    """:obj:`str`, optional: name of the files to write to."""

    std_tolerance_factor: float = -0.01
    """:obj:`float`, optional: standard tolerance with respect to noise.
        Negative for absolute (in eV/Ang).
    """

    min_steps_with_model: int = 0
    """:obj:`int`, optional: minimum steps with model in between ab-initio calls."""

    freeze_hyps: int = 10
    """:obj:`int`, optional: freeze hyperparameters after this many ab-initio calls."""

    write_model: int = 0
    """:obj:`int`, optional: keep at 0 for FLARE++."""




def quicksim(atoms, timestep, steps, calc, otfsettings = OTFSettings(), changes = []):
    """A quick-to-setup simulation using FLARE or FLARE++.

    This function allows the user to do a quick simulation using
    ab-initio calculations and FLARE(++), but at the cost of less flexibility.

    Args:
        atoms (:obj:`Atoms <ase.Atoms>`): atoms object.
        timestep (:obj:`float`): timestep of MD in fs.
        steps (:obj:`int`): amount of MD steps.
        otfsettings (:obj:`OTFSettings`, optional): OTF settings object,
            using defaults if not given.
        changes (optional): array containing settings to update for CHAMP (advanced).
    """
    n_atoms = len(atoms)

    flare_calculator = otfsettings.theory.get_calc(atoms)

    md_engine = 'CustomVerlet'
    md_kwargs = {}

    otf_params = {'init_atoms': None,
                'output_name': otfsettings.output_name,
                'std_tolerance_factor': otfsettings.std_tolerance_factor,
                'max_atoms_added' : n_atoms,
                'force_only' : False,
                'freeze_hyps': otfsettings.freeze_hyps,
                'write_model': otfsettings.write_model,
                'min_steps_with_model': otfsettings.min_steps_with_model,
                'update_style': otfsettings.theory.update_style,
                'update_threshold': otfsettings.theory.update_threshold
                }


    sim = ASE_OTF(atoms,
                timestep = timestep * units.fs,
                number_of_steps = steps,
                dft_calc = calc,
                md_engine = md_engine,
                md_kwargs = md_kwargs,
                update_settings = changes,
                calculator = flare_calculator,
                **otf_params)

    sim.run()



"""Functions and classes for CHAMP I/O."""
from glob import glob
from os import remove
from os.path import exists
from pathlib import Path, PosixPath
from typing import Optional, Type, Union

from pydantic import BaseModel, DirectoryPath, Field, FilePath


class Settings(BaseModel):
    """Data class containing CHAMP configuration.

    This class can hold the neccessery configuration to run CHAMP.
    """
    class General(BaseModel):
        """General module class.
        """

        title: str
        """:obj:`str`: title of the simulation."""

        pool: DirectoryPath = Field(PosixPath('./pool/'), postfix="/")
        """:obj:`path <pathlib.Path>`, optional: location of the pool directory."""

        basis: str
        """:obj:`str`: basis to use (located in pool)."""

        pseudopot: Optional[str]
        """:obj:`str`, optional: pseudopotential to use."""

        mode: str = 'vmc_one_mpi1'
        """:obj:`str`, optional: QMC mode."""

        seed: int = 1837465927472523
        """int, optional: seed for CHAMP."""

        eunit: str = "Hartrees"
        """:obj:`str`, optional: energy units."""

    class Ase(BaseModel):
        """ASE module class.
        """

        iase: int = 1
        iforce_analy: int = 0
        node_cutoff: int = 1
        enode_cutoff: float = 0.05

    class Electrons(BaseModel):
        """Electrons module class.
        """

        nup: int
        """:obj:`int`: amount of electrons with upspin."""

        nelec: int
        """:obj:`int`: total amount of electrons."""

    class Optwf(BaseModel):
        """Wavefunction optimization module class.
        """

        ioptwf: int = 1
        ioptci: int = 1
        ioptjas: int = 1
        ioptorb: int = 1
        nextorb: int = 100
        no_active: int = 1
        nopt_iter: int = 1
        isample_cmat: int = 0
        energy_tol: float = 0.0

    class BlockingVmc(BaseModel):
        """VMC module class.
        """

        vmc_nstep: int = 20
        vmc_nblk: int = 400
        vmc_nblkeq: int = 1
        vmc_nconf_new: int = 0

    class Pseudo(BaseModel):
        """Pseudopotential module class.
        """

        nloc: int = 4
        nquad: int = 6

    general: General
    molecule: Path = Field(prefix="load ")
    """:obj:`path <pathlib.Path>`: path to molecule geometry."""

    basis_num_info: FilePath = Field(prefix="load ")
    """:obj:`path <pathlib.Path>`: path to basis info."""

    determinants: FilePath = Field(prefix="load ")
    """:obj:`path <pathlib.Path>`: path to the determinants file."""

    orbitals: FilePath = Field(prefix="load ")
    """:obj:`path <pathlib.Path>`: path to the orbitals file."""

    jastrow: FilePath = Field(prefix="load ")
    """:obj:`path <pathlib.Path>`: path to the jastrow file."""

    jastrow_der: FilePath = Field(prefix="load ")
    """:obj:`path <pathlib.Path>`: path to the jastrow derivatives file."""

    symmetry: Optional[FilePath] = Field(prefix="load ")
    """optional: path to the symmetry file."""

    ase: Optional[Ase]
    electrons: Electrons
    optwf: Optional[Optwf]
    pseudo: Optional[Pseudo]
    blocking_vmc: Optional[BlockingVmc]

    def write(self, filename='vmc.inp'):
        """Write a this dataclass containing the CHAMP configuration to an input file.

        Args:
            filename (:obj:`str`, optional): input file to write to.
        """

        input_file = open(filename, 'w', encoding="utf-8")

        schema = self.schema()['properties']
        for _, item in enumerate(self):
            if isinstance(item[1], BaseModel):
                schema2 = item[1].schema()['properties']
                input_file.write("\n%module " + item[0] + "\n")
                for _, item2 in enumerate(item[1]):
                    if 'postfix' in schema2[item2[0]]:
                        if item2[1] is not None:
                            input_file.write("\t" + item2[0] + " " + str(item2[1]) +\
                                 schema2[item2[0]]['postfix'] + "\n")
                    else:
                        if item2[1] is not None:
                            input_file.write("\t" + item2[0] + " " + str(item2[1]) + "\n")
                input_file.write("%endmodule\n\n")
            else:
                if 'prefix' in schema[item[0]]:
                    if item[1] is not None:
                        input_file.write(schema[item[0]]['prefix'] + item[0] + " " +\
                             str(item[1]) + '\n')
                else:
                    if item[1] is not None:
                        input_file.write(item[0] + " " + str(item[1]) + '\n')

        input_file.close()

    @classmethod
    def read(cls: Type['BaseModel'], filename: Union[str, Path]) -> 'BaseModel':
        """Read the CHAMP input file and convert it to a dictionary format

        Args:
            filename (:obj:`str`, optional): file of the input file to read and write to.

        Returns:
            :obj:`Settings`: Settings object containing CHAMP configuration.
        """
        # Check if the file exists
        if not exists(filename):
            raise FileNotFoundError(filename + " was not found!")

        path = PosixPath(filename).resolve().parent

        if path.relative_to(path.cwd()).resolve() == path.cwd():
            path =  ""
        else:
            path = path.as_posix() + "/"

        output = {}

        input_file = open(filename, 'r', encoding='utf-8')

        # Set some variables to keep track of modules
        curmod = ""
        inmod = False

        for line in input_file:
            # We skip empty lines
            if not line.strip():
                continue
            # Beginning of a module
            if line.startswith("%module"):
                curmod = line[8:-1]
                inmod = True
                output[curmod] = {}
            # End of a module
            elif line.startswith("%endmodule"):
                inmod = False
            # Loading in a file
            elif line.startswith("load"):
                temp = line[5:-1]
                temp = temp.split()
                key = temp[0]
                temp.pop(0)
                value = ' '.join(temp)
                output[key] = path + value
            # All other tags
            else:
                temp = line[:-1].split()
                key = temp[0]
                temp.pop(0)
                value = ' '.join(temp)
                if inmod:
                    # If we are in a module, add the tag to that subdictionary.
                    if key != "pool":
                        output[curmod][key] = value
                    else:
                        output[curmod][key] = path + value
                else:
                    output[key] = value

        input_file.close()

        return cls(**output)


    def use_opt_wf(self):
        """Function to replace the orbitals, determinants and jastrow with the optimized files.
        """
        opt = ["det_optimal.1.iter*", "orbitals_optimal.1.iter*", "jastrow_optimal.1.iter*"]
        keys = ["determinants", "orbitals", "jastrow"]

        for ind, name in enumerate(opt):
            num = len(glob(name))
            # If there are no optimized WF files, we do not use them
            index = min(num, self.optwf.nopt_iter)
            if num >= index and index > 0:
                opt[ind] = opt[ind].strip('*') + str(index)
                setattr(self, keys[ind], opt[ind])

    def todict(self):
        """Return a dictionary of the class.

        Returns:
            :obj:`str`: json dictionary.
        """
        return self.json(exclude_none=True)

def cleanup(*args):
    """Remove files created by CHAMP.

    This function can cleanup the directory of the simulation.
    Add files as arguments to include (if not in default list)
    or exclude (if in default list) files.

    Args:
        *args: filesnames to include (if not in default list) or exclude (if in default list).
    """
    # Default list of the files that will be removed
    standard = ["force_analytic", "restart_vmc", "output.log", "parser.log", "orbitals_optimal.1.*",
                "jastrow_optimal.1.*", "det_optimal.1.*", "geo_optimal.*", "geo_optimal_final",
                "mc_configs_start", "nohup.out", "vmc_temp.inp"]

    # Add or remove file depending on the input arguments
    for item in args:
        if item in standard:
            standard.remove(item)
        else:
            standard.append(item)

    for file in standard:
        # Take care of the wildcard files
        if file.find('*') != -1:
            for file2 in glob(file):
                remove(file2)
        else:
            if exists(file):
                remove(file)



"""Tools for FLARE and CHAMP simulations."""
import numpy as np

from flare import otf_parser

import matplotlib.pyplot as plt

class Analyze():
    """Tool for analyzing OTF data.

    Args:
        file (Path): otf file to analyze.

    Attributes:
        results (dict): dictionary containing simulation results.
    """

    def __init__(self, file):
        self.otf = otf_parser.OtfAnalysis(file, calculate_energy=True)
        self.results = None

    def to_xyz(self, filename=""):
        """Create an .xyz format file from the OTF trajectory.

        Args:
            filename (str): filename for the .xyz file (include extension).
        """
        if filename == "":
            self.otf.to_xyz('traj.xyz')
        else:
            self.otf.to_xyz(filename)

    def get_data(self):
        """Retrieve the data from the OTF file.

        Note:
            The data will be stores in the results attributes.
            Allowed keys are: times, potential energy, kinetic energy, total energy, temperature.
        """
        dt = self.otf.header['dt']

        frames = np.arange(len(self.otf.times)+1)

        self.results = {}

        self.dft = self.otf.dft_frames
        self.nondft = [i for i in frames if i not in self.dft]

        totE = np.zeros(len(frames))
        kinE = np.zeros(len(frames))
        potE = np.zeros(len(frames))
        temp = np.zeros(len(frames))

        dftframe = 0
        for i in range(len(frames)):
            if i in self.dft:
                totE[i] = self.otf.gp_thermostat['total energy'][dftframe]
                kinE[i] = self.otf.gp_thermostat['kinetic energy'][dftframe]
                potE[i] = self.otf.gp_thermostat['potential energy'][dftframe]
                temp[i] = self.otf.gp_thermostat['temperature'][dftframe]
                dftframe += 1
            else:
                totE[i] = self.otf.thermostat['total energy'][i-1]
                kinE[i] = self.otf.thermostat['kinetic energy'][i-1]
                potE[i] = self.otf.thermostat['potential energy'][i-1]
                temp[i] = self.otf.thermostat['temperature'][i-1]


        self.results['times'] = frames[0:-2] * dt
        self.results['potential energy'] = potE[1:-1]
        self.results['kinetic energy'] = kinE[2:]
        self.results['total energy'] = kinE[2:] + potE[1:-1]
        self.results['temperature'] = temp[2:]

        if max(self.dft) > max(self.nondft):
            self.dft = self.dft[:-1]
        else:
            self.nondft = self.nondft[:-1]

        return self.results

    def plot_energy(self, filename=""):
        """Plot the energy

        Args:
            filename (str): file to save the energy plot to (leave empty if not wanted).
        """
        if self.results is None:
            self.get_data()

        times = self.results['times']
        totE = self.results['total energy']
        potE = self.results['potential energy']
        kinE = self.results['kinetic energy']
        plt.figure()
        plt.grid()
        plt.title("Energy")
        plt.plot(times, totE - totE[0], label="Total Energy")
        plt.plot(times, potE - potE[0], label="Potential Energy")
        plt.plot(times, kinE - kinE[0], label="Kinetic Energy")
        plt.scatter([times[index-1] for index in self.dft[1:]],
            [totE[index-1] - totE[0] for index in self.dft[1:]],
            marker='x', label="QMC Calculations", color='black')
        plt.legend()
        plt.xlabel("Time (ps)")
        plt.ylabel("Energy (eV)")
        if filename != "":
            plt.savefig(filename)

    def plot_temperature(self, filename=""):
        """Plot the temperature

        Args:
            filename (str): file to save the temperature plot to (leave empty if not wanted).
        """
        if self.results is None:
            self.get_data()

            plt.figure()

        times = self.results['times']
        temp = self.results['temperature']
        plt.figure()
        plt.grid()
        plt.title("Temperature")
        plt.plot(times, temp)
        plt.xlabel("Time (ps)")
        plt.ylabel("Temperature (K)")
        if filename != "":
            plt.savefig(filename)



"""FLARE"""
from .otf import C_OTF, C_ASE_OTF
from .utils import Analyze
from .verlet import CustomVerlet
from .quicksim import OTFSettings, quicksim

__all__ = ['C_OTF', 'C_ASE_OTF',
           'Analyze', 'CustomVerlet',
           'OTFSettings', 'quicksim']



"""QMCblip"""
from .champ import CHAMP
from .champio import Settings, cleanup
from .tools import traj_to_db, otf_to_db

__all__ = ['CHAMP', 'Settings',
           'cleanup', 'traj_to_db',
           'otf_to_db']



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
    dependencies = fh.readlines()

setuptools.setup(
    name="QMCblip",
    version="1.0.3",
    author="Emiel Slootman",
    author_email="e.slootman@esciencecenter.nl",
    description="A small package to couple Quantum Monte Carlo codes to Machine Learning Force Fields through ASE.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/NLESC-JCER/QMCblip",
    project_urls={
        "Bug Tracker": "https://github.com/NLESC-JCER/QMCblip/issues",
    },
    license="Apache Software License 2.0",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    test_suite='tests',
    packages=setuptools.find_packages(),
    install_requires=dependencies,
    python_requires=">=3.7, <3.10",
)


import unittest
from qmcblip.champ import CHAMP
from qmcblip.champio import Settings
from ase.calculators.calculator import CalculatorSetupError
from os import remove
from ase import Atoms, units
import numpy as np


class TestChamp(unittest.TestCase):
    def setUp(self):
        self.settings = Settings.read("examples/C2-quicksim/vmc.inp")
        self.atoms = [Atoms('C2', [(0,0,-0.61385), (0,0,0.61385)]),
        Atoms('C4SH4', [(0,0.71495093597,1.31902341514), (0,-0.71495093597,1.31902341514), (0,-1.24055334534,0.05119099544), (0,1.24055334534,0.05119099544), (0,0,-1.15285178278), (0,1.32194923477,2.21441153704), (0,-1.32194923477,2.21441153704), (0,2.27909548764,-0.24288695123), (0,-2.27909548764,-0.24288695123)])]

    def test_setupError(self):
        with self.assertRaises(CalculatorSetupError):
            CHAMP(champ_loc="")
        with self.assertRaises(CalculatorSetupError):
            t = CHAMP(champ_loc="tests/vmc.mov1")
        with self.assertRaises(CalculatorSetupError):
            CHAMP(champ_loc="", vmc_in="examples/C2-quicksim/vmc.inp")
        with self.assertRaises(CalculatorSetupError):
            CHAMP(champ_loc="", settings=self.settings)
            remove('vmc.inp')

    def test_writeInput(self):
        calc = CHAMP(champ_loc="tests/vmc.mov1", vmc_in="examples/C2-quicksim/vmc.inp")
        for atom in self.atoms:
            calc.write_input(atoms=atom)
            f = open('molecule.xyz')
            self.assertEqual(int(f.readline()), len(atom))
            f.readline()
            for i in range(len(atom)):
                a = np.array(f.readline().split()[1:], dtype=float) * units.Bohr/units.Ang
                b = atom[i].position
                for j in range(3):
                    self.assertAlmostEqual(a[j], b[j])
            remove('vmc_temp.inp')
            remove('molecule.xyz')
            f.close()

if __name__ == '__main__':
    unittest.main()